"""Run checks."""

import logging
import signal
import time
import types
from functools import partial
from threading import Event

import click

import dnscheck
from dnscheck.process import domain_checks, nsset_checks, nssets
from dnscheck.settings import APP
from dnscheck.utils import StopProcess, run_unless_stopped

_LOGGER = logging.getLogger(__name__)


def _signal_handler(signum: int, frame: types.FrameType | None, *, stop_event: Event) -> None:  # pragma: no cover
    """Capture signal and set stop event."""
    _LOGGER.info("Captured signal %s, stopping dnscheck daemon", signal.Signals(signum).name)
    stop_event.set()


@click.group()
@click.version_option(version=dnscheck.__version__, message="fred-dnscheck %(version)s")
@click.option("--config", type=click.Path(exists=True, dir_okay=False), help="Config file.")
def cli(config: str) -> None:
    """DNScheck CLI."""
    APP.load(config)


@cli.command()
@click.option(
    "--batch-size", default=1000, show_default=True, help="How many nssets should be processed simultaneously."
)
def sync(batch_size: int) -> None:
    """Sync nssets with domain registry."""
    nssets.sync(batch_size=batch_size)


@cli.command()
@click.option("--loop-delay", default=5, show_default=True, help="Delay in seconds between processing loop runs.")
def run(loop_delay: int) -> None:
    """Run main process loop."""
    _LOGGER.info("Starting dnscheck daemon")
    stop_event = Event()
    signal.signal(signal.SIGTERM, partial(_signal_handler, stop_event=stop_event))

    _run = partial(run_unless_stopped, stop_event=stop_event)
    try:
        while True:
            _run(domain_checks.finish)()
            _run(nsset_checks.evaluate)()
            _run(nsset_checks.start)()
            _run(domain_checks.start)()
            _run(nsset_checks.finish)()
            _run(time.sleep)(loop_delay)
    except KeyboardInterrupt:
        _LOGGER.info("Stopping dnscheck daemon")
    except StopProcess:
        _LOGGER.info("Dnscheck daemon has been gracefully stopped")
