"""Process domain checks."""

import datetime as dt
import logging
from collections import Counter
from typing import cast

from sqlalchemy import func, select
from sqlalchemy.orm import selectinload

from dnscheck.constants import DomainCheckStatus
from dnscheck.db import DomainCheck, NssetCheck, atomic
from dnscheck.schema import Nameserver
from dnscheck.settings import APP

_LOGGER = logging.getLogger(__name__)


def start() -> None:
    """Start multiple domain checks up to limit."""
    with atomic() as sess:
        stmt_count = select(func.count(DomainCheck.id)).where(DomainCheck.status.in_({DomainCheckStatus.STARTED}))
        domain_checks_running = sess.execute(stmt_count).scalar_one()

        if domain_checks_running < APP.settings.max_domain_checks:
            stmt = (
                select(DomainCheck)
                .join(NssetCheck, DomainCheck.nsset_check_id == NssetCheck.id)
                .where(DomainCheck.status == DomainCheckStatus.CREATED)
                .order_by(NssetCheck.started_at.asc())
                .with_for_update(skip_locked=True)
                .limit(APP.settings.max_domain_checks - domain_checks_running)
                .options(selectinload(DomainCheck.nsset_check))
            )
            domain_checks = sess.execute(stmt).scalars().all()

            for check in domain_checks:
                start_one(check)


def start_one(domain_check: DomainCheck) -> None:
    """Start domain check in Zonemaster.

    Args:
        domain_check: DomainCheck instance with status CREATED.
    """
    nameservers = [
        Nameserver(
            fqdn=ns["fqdn"],
            ip_addresses=ns.get("ip_addresses", []),
        )
        for ns in domain_check.nsset_check.nameservers
    ]
    test_id = APP.zonemaster.start_domain_test(
        fqdn=domain_check.domain,
        nameservers=nameservers,
        profile=domain_check.nsset_check.profile,
    )
    domain_check.status = cast(DomainCheckStatus, DomainCheckStatus.STARTED)
    domain_check.test_id = test_id
    domain_check.started_at = dt.datetime.now(tz=dt.UTC)
    _LOGGER.info("Started domain check of %s", domain_check.domain)


def finish() -> None:
    """Finish completed domain checks.

    Find domain checks with state STARTED.
    For each domain check call zonemaster API to check its progress.
    If the test is already finished, fetch the results and finish the domain check.
    """
    with atomic() as db:
        stmt = (
            select(DomainCheck)
            .where(DomainCheck.status == DomainCheckStatus.STARTED)
            .with_for_update(skip_locked=True)
            .options(selectinload(DomainCheck.nsset_check))
        )
        domain_checks = db.execute(stmt).scalars()
        for domain_check in domain_checks:
            progress = APP.zonemaster.test_progress(cast(str, domain_check.test_id))
            if progress >= 100:
                finish_one(domain_check)


def finish_one(domain_check: DomainCheck) -> None:
    """Finish one completed domain check.

    Args:
        domain_check: DomainCheck instance with status STARTED and test progress 100%.
    """
    test_results = APP.zonemaster.get_test_results(cast(str, domain_check.test_id))
    domain_check.status = DomainCheckStatus.FINISHED
    domain_check.finished_at = dt.datetime.now(tz=dt.UTC)
    domain_check.results = test_results
    domain_check.summary = dict(Counter(res["level"] for res in test_results))
    _LOGGER.info("Finished domain check of %s", domain_check.domain)
