"""Process nsset check function."""

import datetime as dt
import logging
from collections.abc import Sequence

from fred_types import NssetId
from regal.exceptions import NssetDoesNotExist
from sqlalchemy import exists, func, select
from sqlalchemy.orm import Session

from dnscheck.constants import DomainCheckStatus, NssetCheckResult, NssetCheckStatus, TestResultLevel
from dnscheck.db import DomainCheck, Nsset, NssetCheck, atomic
from dnscheck.schema import Nameserver
from dnscheck.settings import APP
from dnscheck.utils import async_to_sync

_LOGGER = logging.getLogger(__name__)


def list_candidates(limit: int = 30) -> Sequence[Nsset]:
    """Get nsset check candidates.

    1. nssets that hasn't been checked yet
    2. nssets that hasn't been checked for the longest
       and that were last checked before min_check_interval

    Args:
        limit: Maximum number of returned nssets.
    """
    with atomic(expire_on_commit=False) as db:
        check_datetime_limit = dt.datetime.now(tz=dt.UTC) - APP.settings.min_check_interval
        last_check = (
            select(
                NssetCheck.nsset_id,
                func.max(NssetCheck.started_at).label("started_at"),
            )
            .group_by(NssetCheck.nsset_id)
            .subquery()
        )
        stmt = (
            select(Nsset)
            .outerjoin(last_check, Nsset.nsset_id == last_check.c.nsset_id)
            .outerjoin(
                NssetCheck, (Nsset.nsset_id == NssetCheck.nsset_id) & (NssetCheck.started_at == last_check.c.started_at)
            )
            .where(
                NssetCheck.started_at.is_(None) | NssetCheck.finished_at.is_not(None),  # there is no running check
                last_check.c.started_at.is_(None) | (NssetCheck.finished_at < check_datetime_limit),
            )
            .order_by(last_check.c.started_at.asc().nulls_first())
            .limit(limit)
        )
        return db.execute(stmt).scalars().all()


def start() -> None:
    """Start nsset checks until concurrency limit is reached."""
    with atomic() as sess:
        domain_checks = sess.execute(
            select(func.count(DomainCheck.id)).where(
                DomainCheck.status.in_({DomainCheckStatus.CREATED, DomainCheckStatus.STARTED})
            )
        ).scalar_one()

        if domain_checks >= APP.settings.max_domain_checks:
            return

        for nsset in list_candidates():
            domain_checks += start_one(NssetId(nsset.nsset_id))

            if domain_checks >= APP.settings.max_domain_checks:
                break


@async_to_sync
async def start_one(nsset_id: NssetId) -> int:
    """Start nsset check.

    Args:
        nsset_id: Nsset id to be checked.

    Returns:
        Number of created domain checks.
    """
    try:
        info = await APP.nsset_client.get_nsset_info(nsset_id)
    except NssetDoesNotExist:
        # Nsset has been deleted since last synchronization
        _LOGGER.debug("Nsset %s does not exist anymore. Aborting nsset check.", nsset_id)
        return 0

    # TODO: Get rid of this transformation once fred_types support Pydantic 2 BaseModel.model_dump()
    nameservers = [
        Nameserver(fqdn=dns.fqdn, ip_addresses=[str(ip) for ip in dns.ip_addresses]) for dns in info.dns_hosts
    ]
    with atomic() as sess:
        nsset_check = NssetCheck(
            nsset_id=nsset_id,
            nsset_history_id=info.nsset_history_id,
            handle=info.nsset_handle,
            nameservers=[ns.model_dump() for ns in nameservers],
        )
        sess.add(nsset_check)
        domain_checks = []
        async for domain in APP.domain_client.list_by_nsset(nsset_id):
            domain_checks.append(DomainCheck(domain=domain.fqdn, nsset_check=nsset_check))
        if domain_checks:
            sess.add_all(domain_checks)

        _LOGGER.info("Started nsset check of %s with %s linked domains", nsset_check.nsset_id, len(domain_checks))
        return len(domain_checks)


def evaluate() -> None:
    """Evaluate nsset checks.

    Find nsset checks that don't have any domain checks
    with status STARTED left and evaluate them.
    """
    with atomic() as db:
        stmt = (
            select(NssetCheck)
            .where(
                NssetCheck.status == NssetCheckStatus.STARTED,
                ~exists(DomainCheck.id).where(
                    DomainCheck.nsset_check_id == NssetCheck.id,
                    DomainCheck.status.in_((DomainCheckStatus.CREATED, DomainCheckStatus.STARTED)),
                ),
            )
            .with_for_update(skip_locked=True)
        )
        nsset_checks = db.execute(stmt).scalars()
        for nsset_check in nsset_checks:
            evaluate_one(db, nsset_check)


def evaluate_one(db: Session, nsset_check: NssetCheck) -> None:
    """Evaluate a single nsset check.

    Args:
        db: Active db session.
        nsset_check: Nsset check to be evaluated. All the associated
            domain checks has to be finished.
    """
    stmt = select(DomainCheck.summary).where(DomainCheck.nsset_check_id == nsset_check.id)

    for summary in db.execute(stmt).scalars():
        assert isinstance(summary, dict)  # noqa: S101
        if summary.get(TestResultLevel.CRITICAL, 0) + summary.get(TestResultLevel.ERROR, 0) > 0:
            nsset_check.result = NssetCheckResult.FAIL
            nsset_check.status = NssetCheckStatus.EVALUATED
            break
    else:
        nsset_check.result = NssetCheckResult.SUCCESS
        nsset_check.status = NssetCheckStatus.EVALUATED

    _LOGGER.info("Evaluated nsset check of %s with result %s", nsset_check.nsset_id, nsset_check.result)


def finish() -> None:
    """Finish nsset checks.

    This is a dummy implementation that just closes evaluated nsset checks.
    """
    with atomic() as db:
        stmt = (
            select(NssetCheck).where(NssetCheck.status == NssetCheckStatus.EVALUATED).with_for_update(skip_locked=True)
        )
        nsset_checks = db.execute(stmt).scalars()
        for nsset_check in nsset_checks:
            # TODO: notifications about failed nsset check results
            nsset_check.status = NssetCheckStatus.FINISHED
