"""Process nsset functions."""

import datetime as dt
import logging

from asyncstdlib.itertools import batched
from sqlalchemy import delete, insert, update

from dnscheck.db import Nsset, atomic
from dnscheck.settings import APP
from dnscheck.utils import async_to_sync

_LOGGER = logging.getLogger(__name__)


@async_to_sync
async def sync(batch_size: int = 1000) -> None:
    """Sync nssets with registry.

    Args:
        batch_size: Size of nssets chunk that are processed together.
    """
    _LOGGER.info("Running nsset sync")
    sync_started_at = dt.datetime.now(tz=dt.UTC)

    async for nssets in batched(APP.nsset_client.list(), batch_size):
        with atomic() as sess:
            # First update all the existing nssets
            updated_ids = set(
                sess.execute(
                    update(Nsset)
                    .values(last_synced=dt.datetime.now(tz=dt.UTC))
                    .where(Nsset.nsset_id.in_(n.id for n in nssets))
                    .returning(Nsset.nsset_id)
                ).scalars()
            )
            _LOGGER.info("Updated %s existing nssets", len(updated_ids))

            # Then insert all the new nssets
            new_nssets = [{"nsset_id": n.id, "handle": n.handle} for n in nssets if n.id not in updated_ids]
            if new_nssets:
                sess.execute(insert(Nsset).values(new_nssets))
                _LOGGER.info("Inserted %s new nssets", len(new_nssets))

    with atomic() as sess:
        # Last, delete all nssets that don't exist anymore
        result = sess.execute(delete(Nsset).where(Nsset.last_synced < sync_started_at))
        _LOGGER.info("Deleted %s old nssets", result.rowcount)
