import datetime as dt
from ipaddress import IPv4Address, IPv6Address
from unittest.mock import call

import pytest
from asyncstdlib import iter as aiter
from fred_types import DomainId, DomainRef, NssetHistoryId, NssetId
from pytest_mock import MockFixture
from regal import DnsHost, Nsset as NssetSchema
from regal.exceptions import NssetDoesNotExist
from sqlalchemy import func, select
from time_machine import TimeMachineFixture

from dnscheck.constants import DomainCheckStatus, NssetCheckResult, NssetCheckStatus, TestResultLevel
from dnscheck.db import DomainCheck, NssetCheck, atomic
from dnscheck.process import nsset_checks
from dnscheck.settings import APP
from dnscheck.tests.utils import ModelFactory, override_settings


@override_settings(min_check_interval=dt.timedelta(days=10))
def test_list_nsset_check_candidates(db, time_machine: TimeMachineFixture):
    now = dt.datetime(2000, 1, 1, tzinfo=dt.UTC)
    time_machine.move_to(now)
    dt.datetime.now()  # start the time machine
    mf = ModelFactory()

    with atomic() as sess:
        sess.add_all(
            [
                mf.nsset(1),
                mf.nsset(2),
                mf.nsset(3),
                mf.nsset(4),
                mf.nsset(5),
                mf.nsset(6),
                mf.nsset_check(1, started_at=now - dt.timedelta(days=30), finished_at=now - dt.timedelta(days=29)),
                mf.nsset_check(1, started_at=now - dt.timedelta(days=1)),
                mf.nsset_check(2, started_at=now - dt.timedelta(days=30)),
                mf.nsset_check(3, started_at=now - dt.timedelta(days=2), finished_at=now - dt.timedelta(days=1)),
                mf.nsset_check(4, started_at=now - dt.timedelta(days=30), finished_at=now - dt.timedelta(days=29)),
                mf.nsset_check(5, started_at=now - dt.timedelta(days=20), finished_at=now - dt.timedelta(days=19)),
            ]
        )

    nssets = nsset_checks.list_candidates()

    with atomic() as sess:
        sess.add_all(nssets)
        assert [n.nsset_id for n in nssets] == [
            mf.nsset(6).nsset_id,
            mf.nsset(4).nsset_id,
            mf.nsset(5).nsset_id,
        ]
        # 6 has never been checked
        # 4 has been checked 29 days ago
        # 5 has been checked 19 days ago
        #
        # 1 and 2 have checks that are still running
        # 3 has been checked 1 day ago, which is too recent


class StartNssetChecksTest:
    @pytest.fixture(autouse=True)
    def setup(self, mocker: MockFixture):
        mf = self.mf = ModelFactory()

        self.mock_list = mocker.patch(
            "dnscheck.process.nsset_checks.list_candidates", return_value=[mf.nsset(1), mf.nsset(2), mf.nsset(3)]
        )
        self.mock_start = mocker.patch("dnscheck.process.nsset_checks.start_one", side_effect=[8, 5, 1])

    @pytest.fixture
    def create_domain_checks(self, db):
        with atomic() as sess:
            sess.add_all(
                [
                    self.mf.nsset(42),
                    self.mf.nsset_check(42),
                    self.mf.domain_check("1.example.org"),
                    self.mf.domain_check("2.example.org"),
                ]
            )

    def test_start(self, db):
        with override_settings(max_domain_checks=10):
            nsset_checks.start()

        assert self.mock_list.mock_calls == [call()]
        assert self.mock_start.mock_calls == [call(self.mf.nsset(1).nsset_id), call(self.mf.nsset(2).nsset_id)]

    def test_start_no_candidates(self, db, create_domain_checks, mocker: MockFixture):
        self.mock_list = mocker.patch("dnscheck.process.nsset_checks.list_candidates", return_value=[])

        with override_settings(max_domain_checks=10):
            nsset_checks.start()

        assert self.mock_list.mock_calls == [call()]
        assert self.mock_start.mock_calls == []

    def test_running_checks(self, db, create_domain_checks):
        with override_settings(max_domain_checks=10):
            nsset_checks.start()

        assert self.mock_list.mock_calls == [call()]
        assert self.mock_start.mock_calls == [call(self.mf.nsset(1).nsset_id)]

    def test_too_many_running_checks(self, db, create_domain_checks):
        with override_settings(max_domain_checks=2):
            nsset_checks.start()

        assert self.mock_list.mock_calls == []
        assert self.mock_start.mock_calls == []


def test_start_nsset_check(db, mocker: MockFixture):
    mocker.patch.object(
        APP.nsset_client,
        "get_nsset_info",
        return_value=NssetSchema(
            nsset_id=NssetId("NSS-1"),
            nsset_handle="NSSET",
            nsset_history_id=NssetHistoryId("NSS-2"),
            dns_hosts=[
                DnsHost(fqdn="ns1.example.org", ip_addresses=[]),
                DnsHost(fqdn="ns2.example.org", ip_addresses=[IPv4Address("127.0.0.1"), IPv6Address("::1")]),
            ],
        ),
    )
    mocker.patch.object(
        APP.domain_client,
        "list_by_nsset",
        return_value=aiter(
            [
                DomainRef(id=DomainId("DOMAIN-1"), fqdn="domain1.example.org"),
                DomainRef(id=DomainId("DOMAIN-2"), fqdn="domain2.example.org"),
                DomainRef(id=DomainId("DOMAIN-3"), fqdn="domain3.example.org"),
            ]
        ),
    )

    nsset_checks.start_one(NssetId("NSS-1"))

    with atomic() as sess:
        check = sess.execute(select(NssetCheck)).scalar_one()
        assert check.nsset_id == "NSS-1"
        assert check.nsset_history_id == "NSS-2"
        assert check.handle == "NSSET"
        assert check.nameservers == [
            {"fqdn": "ns1.example.org", "ip_addresses": []},
            {"fqdn": "ns2.example.org", "ip_addresses": ["127.0.0.1", "::1"]},
        ]
        domain_checks = sess.execute(select(DomainCheck)).scalars().all()
        assert [dc.domain for dc in domain_checks] == [
            "domain1.example.org",
            "domain2.example.org",
            "domain3.example.org",
        ]
        assert all(dc.nsset_check == check for dc in domain_checks)


def test_start_nsset_check_no_domains(db, mocker: MockFixture):
    mocker.patch.object(
        APP.nsset_client,
        "get_nsset_info",
        return_value=NssetSchema(
            nsset_id=NssetId("NSS-1"),
            nsset_handle="NSSET",
            nsset_history_id=NssetHistoryId("NSS-2"),
            dns_hosts=[
                DnsHost(fqdn="ns1.example.org", ip_addresses=[]),
                DnsHost(fqdn="ns2.example.org", ip_addresses=[IPv4Address("127.0.0.1"), IPv6Address("::1")]),
            ],
        ),
    )
    mocker.patch.object(APP.domain_client, "list_by_nsset", return_value=aiter([]))

    nsset_checks.start_one(NssetId("NSS-1"))

    with atomic() as sess:
        assert sess.execute(select(func.count(NssetCheck.id))).scalar_one() == 1
        assert sess.execute(select(func.count(DomainCheck.id))).scalar_one() == 0


def test_start_nsset_check_does_not_exist(db, mocker: MockFixture):
    mocker.patch.object(APP.nsset_client, "get_nsset_info", side_effect=NssetDoesNotExist)

    nsset_checks.start_one(NssetId("NSS-1"))

    with atomic() as sess:
        assert sess.execute(select(func.count(NssetCheck.id))).scalar_one() == 0
        assert sess.execute(select(func.count(DomainCheck.id))).scalar_one() == 0


def test_evaluate(db):
    mf = ModelFactory()

    with atomic() as sess:
        sess.add_all(
            [
                mf.nsset(1),
                nsset_check1 := mf.nsset_check(1, status=NssetCheckStatus.STARTED),
                mf.domain_check(
                    "01.example.com", status=DomainCheckStatus.FINISHED, summary={TestResultLevel.CRITICAL.value: 1}
                ),
                mf.domain_check(
                    "02.example.com",
                    status=DomainCheckStatus.FINISHED,
                    summary={TestResultLevel.ERROR.value: 1, TestResultLevel.WARNING: 2},
                ),
                nsset_check2 := mf.nsset_check(2, status=NssetCheckStatus.STARTED),
                mf.domain_check(
                    "11.example.com", status=DomainCheckStatus.FINISHED, summary={TestResultLevel.INFO.value: 1}
                ),
                mf.domain_check(
                    "12.example.com",
                    status=DomainCheckStatus.FINISHED,
                    summary={TestResultLevel.NOTICE.value: 1, TestResultLevel.WARNING: 2},
                ),
                nsset_check3 := mf.nsset_check(3, status=NssetCheckStatus.STARTED),
                mf.domain_check(
                    "21.example.com",
                    status=DomainCheckStatus.STARTED,
                ),
                nsset_check4 := mf.nsset_check(4, status=NssetCheckStatus.STARTED),
            ]
        )

    nsset_checks.evaluate()

    with atomic() as sess:
        sess.add_all([nsset_check1, nsset_check2, nsset_check3, nsset_check4])

        assert nsset_check1.status == NssetCheckStatus.EVALUATED
        assert nsset_check1.result == NssetCheckResult.FAIL

        assert nsset_check2.status == NssetCheckStatus.EVALUATED
        assert nsset_check2.result == NssetCheckResult.SUCCESS

        assert nsset_check3.status == NssetCheckStatus.STARTED
        assert nsset_check3.result is None

        assert nsset_check4.status == NssetCheckStatus.EVALUATED
        assert nsset_check4.result == NssetCheckResult.SUCCESS


def test_finish(db):
    mf = ModelFactory()

    with atomic() as sess:
        sess.add_all(
            [
                nsset_check1 := mf.nsset_check(1, status=NssetCheckStatus.STARTED),
                nsset_check2 := mf.nsset_check(1, status=NssetCheckStatus.EVALUATED),
                nsset_check3 := mf.nsset_check(1, status=NssetCheckStatus.EVALUATED),
                nsset_check4 := mf.nsset_check(1, status=NssetCheckStatus.FINISHED),
            ]
        )

    nsset_checks.finish()

    with atomic() as sess:
        sess.add_all([nsset_check1, nsset_check2, nsset_check3, nsset_check4])

        assert nsset_check1.status == NssetCheckStatus.STARTED
        assert nsset_check2.status == NssetCheckStatus.FINISHED
        assert nsset_check3.status == NssetCheckStatus.FINISHED
        assert nsset_check4.status == NssetCheckStatus.FINISHED
