from unittest.mock import call, sentinel

import pytest
from pydantic import BaseModel, ValidationError
from pytest_mock import MockerFixture

import dnscheck.settings
from dnscheck.settings import APP, HttpUrlString
from dnscheck.zonemaster import ZonemasterClient

from .utils import override_settings


@override_settings(zonemaster_api_url="http://zonemaster.example.com/api/")
def test_app_zonemaster(mocker: MockerFixture):
    spy = mocker.spy(dnscheck.settings, "ZonemasterClient")
    zonemaster = APP.zonemaster
    assert zonemaster is APP.zonemaster
    assert isinstance(zonemaster, ZonemasterClient)
    assert spy.mock_calls == [call("http://zonemaster.example.com/api/")]


@pytest.mark.parametrize("client,attr", [("DomainClient", "domain_client"), ("NssetClient", "nsset_client")])
def test_app_registry_client(mocker: MockerFixture, client: str, attr: str):
    mocker.patch.object(APP.settings, "registry_netloc", sentinel.netloc)
    mocker.patch.object(APP.settings, "registry_ssl_cert", sentinel.ssl_cert)
    client_mock = mocker.patch(f"dnscheck.settings.{client}", return_value=sentinel.client)
    cred_mock = mocker.patch("dnscheck.settings.make_credentials", return_value=sentinel.credentials)

    assert getattr(APP, attr) == sentinel.client

    assert client_mock.mock_calls == [call(sentinel.netloc, sentinel.credentials)]
    assert cred_mock.mock_calls == [call(sentinel.ssl_cert)]


def test_app_clear_cache():
    """Test that all cached properties are cleared."""
    zonemaster = APP.zonemaster
    domain_client = APP.domain_client
    nsset_client = APP.nsset_client

    APP.clear_cache()

    assert APP.zonemaster is not zonemaster
    assert APP.domain_client is not domain_client
    assert APP.nsset_client is not nsset_client


class UrlModel(BaseModel):
    url: HttpUrlString


@pytest.mark.parametrize("url", ["https://example.com/api", "http://example.com/"])
def test_http_url_string_valid(url: str):
    assert UrlModel(url=url).url == url


@pytest.mark.parametrize("url", ["gazpacho!", "", "example.com"])
def test_http_url_string_invalid(url: str):
    with pytest.raises(ValidationError):
        UrlModel(url=url)
