import pytest
from httpx import Response
from respx.router import MockRouter

import dnscheck
from dnscheck.schema import Nameserver
from dnscheck.zonemaster import ZonemasterClient

API_URL = "http://example.com/api/"


@pytest.fixture
def zonemaster_client() -> ZonemasterClient:
    """Provide dummy zonemaster client."""
    return ZonemasterClient(API_URL)


def test_start_domain_test_simple(respx_mock: MockRouter, zonemaster_client):
    """Test start_domain_test without profile or nameservers."""
    request_data = {
        "jsonrpc": "2.0",
        "id": 1,
        "method": "start_domain_test",
        "params": {
            "client_id": "fred-dnscheck",
            "domain": "example.com",
            "client_version": dnscheck.__version__,
            "ipv6": True,
            "ipv4": True,
            "nameservers": [],
        },
    }
    respx_mock.post(API_URL, json=request_data).mock(
        return_value=Response(200, json={"jsonrpc": "2.0", "id": 1, "result": "gazpacho"})
    )

    assert zonemaster_client.start_domain_test(fqdn="example.com", nameservers=[]) == "gazpacho"


def test_start_domain_test_full(respx_mock: MockRouter, zonemaster_client):
    """Test start_domain_test with profile and nameservers."""
    request_data = {
        "jsonrpc": "2.0",
        "id": 1,
        "method": "start_domain_test",
        "params": {
            "client_id": "fred-dnscheck",
            "domain": "example.com",
            "client_version": dnscheck.__version__,
            "ipv6": True,
            "ipv4": True,
            "profile": "default",
            "nameservers": [
                {"ns": "ns1.example.com"},
                {"ns": "ns2.example.com", "ip": "10.0.0.1"},
                {"ns": "ns2.example.com", "ip": "10.0.0.2"},
            ],
        },
    }

    respx_mock.post(API_URL, json=request_data).mock(
        return_value=Response(200, json={"jsonrpc": "2.0", "id": 1, "result": "gazpacho"})
    )

    assert (
        zonemaster_client.start_domain_test(
            fqdn="example.com",
            nameservers=[
                Nameserver(fqdn="ns1.example.com"),
                Nameserver(fqdn="ns2.example.com", ip_addresses=["10.0.0.1", "10.0.0.2"]),
            ],
            profile="default",
        )
        == "gazpacho"
    )


def test_test_progress(respx_mock: MockRouter, zonemaster_client):
    request_data = {
        "jsonrpc": "2.0",
        "id": 1,
        "method": "test_progress",
        "params": {"test_id": "gazpacho"},
    }
    respx_mock.post(API_URL, json=request_data).mock(
        return_value=Response(200, json={"jsonrpc": "2.0", "id": 1, "result": 42})
    )

    assert zonemaster_client.test_progress("gazpacho") == 42


def test_get_test_results(respx_mock: MockRouter, zonemaster_client):
    request_data = {
        "jsonrpc": "2.0",
        "id": 1,
        "method": "get_test_results",
        "params": {"id": "gazpacho", "language": "en"},
    }
    respx_mock.post(API_URL, json=request_data).mock(
        return_value=Response(
            200,
            json={
                "jsonrpc": "2.0",
                "id": 1,
                "result": {
                    "params": {},
                    "results": [
                        {
                            "module": "SYSTEM",
                            "message": "Using version v1.0.14 of the Zonemaster engine.\n",
                            "level": "INFO",
                        },
                    ],
                },
            },
        )
    )

    assert zonemaster_client.get_test_results("gazpacho") == [
        {
            "module": "SYSTEM",
            "message": "Using version v1.0.14 of the Zonemaster engine.\n",
            "level": "INFO",
        },
    ]
