"""DNScheck utils."""

import asyncio
from collections.abc import Callable, Coroutine
from functools import wraps
from threading import Event
from typing import ParamSpec, TypeVar

P = ParamSpec("P")
T = TypeVar("T")


def async_to_sync(func: Callable[P, Coroutine[None, None, T]]) -> Callable[P, T]:
    """Run async function.

    This function can only be called from synchronous context.
    If you're in asynchronous context, use `await` directly.
    """

    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return asyncio.run(func(*args, **kwargs))

    return wrapper


class StopProcess(Exception):
    """Stop processing."""


def run_unless_stopped(func: Callable[P, T], *, stop_event: Event) -> Callable[P, T]:
    """Run function unless stop event has been triggered."""

    def _wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        if stop_event.is_set():
            raise StopProcess
        else:
            return func(*args, **kwargs)

    return _wrapper


__all__ = [
    "async_to_sync",
    "run_unless_stopped",
    "StopProcess",
]
